"""
    QuaO Project aws_braket_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from ..device.device import Device
from ..provider.provider import Provider
from ...enum.job_status import JobStatus
from ...util.json_parser_util import JsonParserUtils
from ...config.logging_config import *
from dateutil.parser import parse


class AwsBraketDevice(Device):
    def _get_name(self) -> str:
        return str(self.device.name)

    def __init__(self, provider: Provider,
                 device_specification: str,
                 s3_bucket_name: str,
                 s3_prefix: str):
        super().__init__(provider, device_specification)
        self.s3_folder = (s3_bucket_name, s3_prefix)

    def _create_job(self, circuit, shots):
        logger.debug('Create AWS Braket job with {0} shots'.format(shots))

        job = self.device.run(task_specification=circuit,
                              s3_destination_folder=self.s3_folder,
                              shots=shots)
        return job

    def _is_simulator(self) -> bool:
        return 'SIMULATOR'.__eq__(self.device.type.value)

    def _produce_histogram_data(self, job_result) -> dict:
        return dict(job_result.measurement_counts)

    def _get_provider_job_id(self, job) -> str:
        return job.id

    def _get_job_status(self, job) -> str:
        job_state = job.state()
        if JobStatus.COMPLETED.value.__eq__(job_state):
            job_state = JobStatus.DONE.value
        return job_state


    def _calculate_execution_time(self, job_result):
        if 'task_metadata' not in job_result:
            return

        task_metadata = job_result['task_metadata']

        if task_metadata is None \
                or not bool(task_metadata)\
                or 'createdAt' not in task_metadata \
                or 'endedAt' not in task_metadata:
            return

        created_at = task_metadata['createdAt']
        ended_at = task_metadata['endedAt']

        if created_at is None or ended_at is None:
            return

        created_at = parse(created_at.replace("T", " ").replace("Z", ""))
        ended_at = parse(ended_at.replace("T", " ").replace("Z", ""))

        offset = ended_at - created_at

        self.execution_time = offset.total_seconds()



"""
    QuaO Project provider_factory.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from ..enum.provider_type import ProviderType
from ..model.provider.aws_braket_provider import AwsBraketProvider
from ..model.provider.ibm_cloud_provider import IbmCloudProvider
from ..model.provider.ibm_quantum_provider import IbmQuantumProvider
from ..model.provider.quao_provider import QuaoProvider


class ProviderFactory:

    @staticmethod
    def create_provider(provider_type: str, authentication: dict):
        """

        @param provider_type:
        @param authentication:
        @return:
        """

        if ProviderType.QUAO_QUANTUM_SIMULATOR.value.__eq__(provider_type):
            return QuaoProvider()

        if ProviderType.IBM_QUANTUM.value.__eq__(provider_type):
            return IbmQuantumProvider(authentication.get('token'))

        if ProviderType.IBM_CLOUD.value.__eq__(provider_type):
            return IbmCloudProvider(authentication.get('token'), authentication.get('crn'))

        if ProviderType.AWS_BRAKET.value.__eq__(provider_type):
            return AwsBraketProvider(
                authentication.get('accessKey'),
                authentication.get('secretKey'),
                authentication.get('regionName'))

        raise Exception("Unsupported provider!")



"""
    QuaO Project invocation_handler.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

from .backend import Backend
from .data.request.request_data import RequestData
from .util.response_utils import ResponseUtils


class InvocationHandler:
    def __init__(self, event):
        self.event = event

    def invoke(self, circuit_preparation, post_processing):
        """

        @param post_processing:
        @param circuit_preparation:
        @return:
        """

        request_data = RequestData(self.event)
        backend = Backend(request_data)

        circuit = circuit_preparation(request_data.input)

        job = backend.submit_job(circuit, post_processing)

        response = ResponseUtils.generate_response(job)

        return response



"""
    QuaO Project job_status.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from qiskit_ibm_runtime import QiskitRuntimeService

from ...data.job.job_response import JobResponse
from ...enum.job_status import JobStatus
from ...enum.media_type import MediaType
from ...util.json_parser_util import JsonParserUtils
from ...config.logging_config import *


class Job:
    def __init__(self, channel, token, crn, provider_job_id, user_identity):
        self.channel = channel if channel else "ibm_quantum"
        self.token = token if token else ""
        self.crn = crn
        self.job_id = provider_job_id if provider_job_id else ""
        self.retrieve_job = self.__get_job()
        self.user_identity = user_identity

    def fetch_job(self) -> JobResponse:
        logger.info("Start get job info.")

        provider_job_id = self._get_provider_job_id()
        job_status = self._get_job_status()
        job_result_dictionary = {}
        content_type = None
        job_histogram = None
        execution_time = None

        try:
            if JobStatus.DONE.value.__eq__(job_status):
                logger.info("Try to get job result")

                job_result = self.retrieve_job.result()
                job_status = self._get_job_status()

                logger.info("Parsing job result")
                job_result_dictionary = self._parse_job_result(job_result)

                content_type = MediaType.APPLICATION_JSON.value

                logger.info("Producing histogram")
                job_histogram = self._produce_histogram_data(job_result)

                logger.debug('Calculating execution time ....')
                execution_time = self._get_execution_time(job_result_dictionary)
                logger.debug('Execution time calculation completed!')

        except Exception as exception:

            logger.info("Try to job result error.")
            job_result_dictionary = {
                "error": "Exception when invoke job on device: " + self.retrieve_job.backend().backend_name,
                "exception": str(exception)
            }

            job_status = JobStatus.ERROR.value

        return JobResponse(
            provider_job_id=provider_job_id,
            job_status=job_status,
            job_result=job_result_dictionary,
            content_type=content_type,
            job_histogram=job_histogram,
            user_identity=self.user_identity,
            execution_time=execution_time
        )

    def __get_job(self):
        logger.info("Has IBM token: {0}".format((self.token is not None) and (len(self.token) > 0)))

        service = QiskitRuntimeService(channel=self.channel,
                                       token=self.token,
                                       instance=self.crn)

        return service.job(job_id=self.job_id)

    def _get_provider_job_id(self) -> str:
        return self.retrieve_job.job_id()

    def _get_job_status(self) -> str:
        return self.retrieve_job.status().name

    @staticmethod
    def _produce_histogram_data(job_result) -> dict:
        try:
            histogram_data = job_result.get_counts()
        except Exception:
            histogram_data = None

        return histogram_data

    def _parse_job_result(self, job_result):
        if self.retrieve_job.program_id == 'sampler':
            return JsonParserUtils.parse(job_result.__dict__)
        return JsonParserUtils.parse(job_result.to_dict())

    @staticmethod
    def _get_execution_time(job_result):
        if 'metadata' not in job_result:
            return None

        metadata = job_result['metadata']

        if metadata is None \
                or not bool(metadata) \
                or 'time_taken_execute' not in metadata:
            return None

        return metadata['time_taken_execute']
